/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define __NO_VERSION__

#include "xfs_locl.h"
#include "xfs_dev.h"
#include "xfs_cache.h"
#include "xfs_dirent.h"
#include "xfs_syscalls.h"
#include <linux/binfmts.h>
#include <linux/file.h>

RCSID("$Id: xfs_inodeops.c,v 1.19 1998/03/25 04:19:19 assar Exp $");

static int
xfs_lookup (struct inode * dir, const char * name, int len,
	    struct inode ** result);

static int
xfs_fsync(struct inode *, struct file *);

static int
xfs_open(struct inode *i, struct file *f);

int
xfs_permission(struct inode *inode, int mode);

static int
xfs_read_file(struct inode *inode, struct file *file, char *buf, int count);

static int
xfs_write_file(struct inode *inode, struct file *file, const char *buf, int count);

static int
xfs_create (struct inode * dir, const char * name, int len, int mode, struct inode ** result);

static int
xfs_readdir(struct inode * inode, struct file * file,
	    void * dirent, filldir_t filldir);

int
xfs_rename (struct inode * old_dir, const char * old_name, int old_len,
	     struct inode * new_dir, const char * new_name, int new_len,
	     int must_be_dir);

int
xfs_unlink (struct inode * dir, const char * name, int len);

static int
xfs_readlink(struct inode *inode, char *buffer, int buflen);

static int
xfs_follow_link(struct inode * dir, struct inode * inode,
		int flag, int mode, struct inode ** res_inode);

static void
xfs_release_file (struct inode *inode, struct file *file);

static struct file_operations xfs_file_operations = {
        NULL,                   /* lseek */
        xfs_read_file,          /* read */
        xfs_write_file,         /* write */
        NULL,                   /* readdir */
        NULL,                   /* select */
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        xfs_open,               /* open */
        xfs_release_file,       /* release */
        xfs_fsync,              /* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

struct inode_operations xfs_file_inode_operations = {
        &xfs_file_operations,   /* default file operations */
        NULL,                   /* create */
        NULL,                   /* lookup */
        NULL,                   /* link */
        NULL,                   /* unlink */
        NULL,                   /* symlink */
        NULL,                   /* mkdir */
        NULL,                   /* rmdir */
        NULL,                   /* mknod */
        NULL,                   /* rename */
        NULL,                   /* readlink */
        NULL,                   /* follow_link */
        generic_readpage,       /* readpage */
        NULL,                   /* writepage */
        NULL,                   /* bmap */
        NULL,                   /* truncate (not yet) */
        xfs_permission,         /* permission */
        NULL                    /* smap */
};

static struct file_operations xfs_dead_operations = {
        NULL,                   /* lseek */
        NULL,                   /* read */
        NULL,                   /* write */
        NULL,                   /* readdir */
        NULL,                   /* select */
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        NULL,                   /* open */
        NULL,                   /* release */
        NULL,                   /* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

struct inode_operations xfs_dead_inode_operations = {
        &xfs_dead_operations,   /* default file operations */
        NULL,                   /* create */
        NULL,                   /* lookup */
        NULL,                   /* link */
        NULL,                   /* unlink */
        NULL,                   /* symlink */
        NULL,                   /* mkdir */
        NULL,                   /* rmdir */
        NULL,                   /* mknod */
        NULL,                   /* rename */
        NULL,                   /* readlink */
        NULL,                   /* follow_link */
        NULL,                   /* readpage */
        NULL,                   /* writepage */
        NULL,                   /* bmap */
        NULL,                   /* truncate */
        NULL,                   /* permission */
        NULL                    /* smap */
};

static struct file_operations xfs_dir_operations = {
        NULL,                   /* lseek */
        NULL,                   /* read */
        NULL,                   /* write */
        xfs_readdir,            /* readdir */
        NULL,                   /* select */
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        NULL,                   /* open */
        NULL,                   /* release */
        NULL,                   /* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

struct inode_operations xfs_dir_inode_operations = {
        &xfs_dir_operations,   /* default file operations */
        xfs_create,             /* create */
        xfs_lookup,             /* lookup */
        NULL,                   /* link (not yet) */
        xfs_unlink,             /* unlink */
        NULL,                   /* symlink (not yet) */
        NULL,                   /* mkdir (not yet) */
        NULL,                   /* rmdir (not yet) */
        NULL,                   /* mknod */
        xfs_rename,             /* rename */
        NULL,                   /* readlink */
        NULL,                   /* follow_link */
        NULL,                   /* readpage */
        NULL,                   /* writepage */
        NULL,                   /* bmap */
        NULL,                   /* truncate */
        xfs_permission,         /* permission */
        NULL                    /* smap */
};

static struct file_operations xfs_link_operations = {
        NULL,                   /* lseek */
        NULL,                   /* read */
        NULL,                   /* write */
        NULL,                   /* readdir */
        NULL,                   /* select */
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        NULL,                   /* open */
        NULL,                   /* release */
        NULL,                   /* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

struct inode_operations xfs_link_inode_operations = {
        &xfs_link_operations,   /* default file operations */
        NULL,                   /* create */
        NULL,                   /* lookup */
        NULL,                   /* link */
        NULL,                   /* unlink */
        NULL,                   /* symlink */
        NULL,                   /* mkdir */
        NULL,                   /* rmdir */
        NULL,                   /* mknod */
        NULL,                   /* rename */
        xfs_readlink,           /* readlink */
        xfs_follow_link,        /* follow_link */
        NULL,                   /* readpage */
        NULL,                   /* writepage */
        NULL,                   /* bmap */
        NULL,                   /* truncate */
        NULL,                   /* permission */
        NULL                    /* smap */
};

int xfs_lookup (struct inode * dir, const char * name, int len,
		 struct inode ** result)
{
    struct xfs_message_getnode msg;
    struct xfs *xfsp;
    int error = 0;
    
    struct xfs_node *d;
    struct xfs_node *x;
    
    XFSDEB(XDEBVNOPS, ("xfs_lookup: name: %s(%d) ", name, len));
    
    if (dir == NULL) {
	XFSDEB(XDEBVNOPS, ("ENOENT\n"));
	return -ENOENT;
    }
    if (!S_ISDIR(dir->i_mode)) {
	iput(dir);
	XFSDEB(XDEBVNOPS, ("ENOTDIR\n"));
	return -ENOTDIR;
    }
    xfsp = XFS_FROM_VNODE(dir);
    d = VNODE_TO_XNODE(dir);
    if (d == NULL || xfsp == NULL) {
	iput(dir);
	XFSDEB(XDEBVNOPS, ("ENODEV\n"));
	return -ENODEV;
    }

    *result = NULL;
    do {
#ifdef notdef_but_correct_and_not_ported
	error = xfs_access(dvp, VEXEC, cred);
	if (error != 0)
	    goto done;
#endif

	if ((x = xfs_cache_lookup(VNODE_TO_XNODE(dir), name, len)) == NULL) {
	    msg.header.opcode = XFS_MSG_GETNODE;

	    msg.cred.uid = current->uid;
	    msg.cred.pag = xfs_get_pag();
	    msg.parent_handle = d->handle;
	    
	    memcpy(msg.name, name, len);
	    msg.name[len]='\0';
	    
	    error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
	    
	    if (error == 0)
		error = ((struct xfs_message_wakeup *) &msg)->error;
	} else {
	    *result=iget(dir->i_sb,(ino_t) x);
	    goto done;
	}
    } while (error == 0);

done:
    if (error)
	XFSDEB(XDEBVNOPS, ("error %d", error));
    XFSDEB(XDEBVNOPS, ("\n"));
    iput(dir);
    return -error;
}

static int
xfs_open_valid(struct inode *vp, u_int tok)
{
  struct xfs *xfsp = XFS_FROM_VNODE(vp);
  struct xfs_node *xn = VNODE_TO_XNODE(vp);
  int error = 0;

  XFSDEB(XDEBVFOPS, ("xfs_open_valid\n"));
  
  do {
    if (!XFS_TOKEN_GOT(xn, tok))
      {
	struct xfs_message_open msg;
	msg.header.opcode = XFS_MSG_OPEN;
	msg.cred.uid = current->uid;
	msg.cred.pag = xfs_get_pag();
	msg.handle = xn->handle;
	msg.tokens = tok;
	error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
	if (error == 0)
	  error = ((struct xfs_message_wakeup *) &msg)->error;
      }
    else
      {
	goto done;
      }
  } while (error == 0);

 done:
  return -error;
}

static int
xfs_open(struct inode *i, struct file *f)
{
  XFSDEB(XDEBVNOPS, ("xfs_open f->f_mode: %d\n",f->f_mode));

  if (f->f_mode & FMODE_WRITE)
    return xfs_open_valid(i, XFS_OPEN_NW);
  else
    return xfs_open_valid(i, XFS_OPEN_NR);
}

static int
do_fsync(struct xfs *xfsp, struct xfs_node *xn, u_int flag)
{
    int error;
    struct xfs_message_putdata msg;

    msg.header.opcode = XFS_MSG_PUTDATA;
    msg.cred.uid = current->uid;
    msg.cred.pag = xfs_get_pag();
    msg.handle = xn->handle;
    msg.flag = flag;

    error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (error == 0)
	error = ((struct xfs_message_wakeup *) & msg)->error;

    if (error == 0)
	xn->flags &= ~XFS_DATA_DIRTY;

    return error;
}

static int
xfs_fsync(struct inode *inode, struct file *file)
{
    struct xfs *xfsp = XFS_FROM_VNODE(inode);
    struct xfs_node *xn = VNODE_TO_XNODE(inode);
    int error = 0;

    XFSDEB(XDEBVNOPS, ("xfs_fsync: 0x%x\n", (int) inode));

    if (xn == NULL)
	return 0;

    if (xn->flags & XFS_DATA_DIRTY)
	error = do_fsync(xfsp, xn, XFS_DATA_DIRTY);
    return error;
}

static int
xfs_attr_valid(struct inode * vp, u_int tok)
{
    struct xfs *xfsp = XFS_FROM_VNODE(vp);
    struct xfs_node *xn = VNODE_TO_XNODE(vp);
    int error = 0;

    pag_t pag = xfs_get_pag();

    do {
        if (!XFS_TOKEN_GOT(xn, tok)) {
            struct xfs_message_getattr msg;

            msg.header.opcode = XFS_MSG_GETATTR;
            msg.cred.uid = current->uid;
            msg.cred.pag = pag;
            msg.handle = xn->handle;
            error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
            if (error == 0)
                error = ((struct xfs_message_wakeup *) & msg)->error;
        } else {
            goto done;
        }
    } while (error == 0);

done:
    return error;
}


int
xfs_permission(struct inode *inode, int mode)
{
    int error = 0;
    pag_t pag = xfs_get_pag();
    
    XFSDEB(XDEBVNOPS, ("xfs_access mode = 0%o\n", mode));

    error = xfs_attr_valid(inode, XFS_ATTR_R);
    if (error == 0) {
	struct xfs_node *xn = VNODE_TO_XNODE(inode);
	int i;

	error = EACCES;

	for (i = 0; i < MAXRIGHTS; i++) {
	    /* XXX locking should be done */
	    if (xn->id[i] == pag) {
		error = 0;
		if (mode & MAY_READ) {
		    if ((xn->rights[i] & XFS_RIGHT_R) == 0) {
			error = EACCES;
		    }
		}
		if (mode & MAY_WRITE) {
		    if ((xn->rights[i] & XFS_RIGHT_W) == 0) {
			error = EACCES;
		    }
		}
		if (mode & MAY_EXEC) {
		    if ((xn->rights[i] & XFS_RIGHT_X) == 0) {
			error = EACCES;
		    }
		}
		break;
	    }
	}
    }
    XFSDEB(XDEBVNOPS, ("xfs_access(0%o) = %d\n", mode, error));
    return -error;
}

static int
xfs_data_valid(struct inode *vp, u_int tok)
{
  struct xfs *xfsp = XFS_FROM_VNODE(vp);
  struct xfs_node *xn = VNODE_TO_XNODE(vp);
  int error = 0;
  
  do {
    if (!XFS_TOKEN_GOT(xn, tok))
      {
	struct xfs_message_getdata msg;
	msg.header.opcode = XFS_MSG_GETDATA;
	msg.cred.uid = current->uid;
	msg.cred.pag = xfs_get_pag();
	msg.handle = xn->handle;
	msg.tokens = tok;
	error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
	if (error == 0)
	  error = ((struct xfs_message_wakeup *) &msg)->error;
      }
    else
      {
	goto done;
      }
  } while (error == 0);

 done:
  return -error;
}


static int xfs_read_file(struct inode *inode, struct file *file, char *buf, int count)
{
  int error = 0;

  XFSDEB(XDEBVNOPS, ("xfs_read_file file->f_pos: %d count: %d\n",(int) file->f_pos,count));

  error = xfs_data_valid(inode, XFS_DATA_R);

  if (error == 0) {
    struct inode *t = DATA_FROM_VNODE(inode);
    t->i_count++;
    if (t->i_op &&
	t->i_op->default_file_ops &&
	t->i_op->default_file_ops->read)
      error = t->i_op->default_file_ops->read(t, file, buf, count);
    else {	
	printk("XFS Panic! read not defined in cache file system\n");
	error = -EINVAL;
    }
    iput(t);
  }
  return error;
}

static int xfs_write_file(struct inode *inode, struct file *file, const char *buf, int count)
{
    int error = 0;
    
    XFSDEB(XDEBVNOPS, ("xfs_write_file file->f_pos: %d count: %d\n",(int) file->f_pos,count));
    
    error = xfs_data_valid(inode, XFS_DATA_W);
    
    if (error == 0) {
	struct inode *t = DATA_FROM_VNODE(inode);
	t->i_count++;
	if (t->i_op &&
	    t->i_op->default_file_ops &&
	    t->i_op->default_file_ops->read) {
	    XFSDEB(XDEBVNOPS, ("xfs_write_file t:%p",t));
	    error = t->i_op->default_file_ops->write(t, file, buf, count);
	    XFSDEB(XDEBVNOPS, ("xfs_write_file error: %d\n",error));
	    VNODE_TO_XNODE(inode)->flags |= XFS_DATA_DIRTY;
	} else {	
	    printk("XFS Panic! write not defined in cache file system\n");
	    error = -EINVAL;
	}
	iput(t);
    }
    return error;
}


static int
xfs_create (struct inode * dir, const char * name, int len, int mode, struct inode ** result)
{
    struct xfs *xfsp;
    struct xfs_node *xn;
    int error = 0;
    struct xfs_message_create msg;

    *result = NULL;
    if (!dir)
	return -ENOENT;

    xfsp = XFS_FROM_VNODE(dir);
    xn = VNODE_TO_XNODE(dir);

    XFSDEB(XDEBVNOPS, ("xfs_create: (%s, %d)\n",
		       name, len));

    msg.header.opcode = XFS_MSG_CREATE;
    msg.parent_handle = xn->handle;
    strncpy(msg.name, name, 256);
    XA_CLEAR(&msg.attr);
    XA_SET_MODE(&msg.attr, mode);
    XA_SET_UID(&msg.attr, current->uid);
    XA_SET_GID(&msg.attr, current->gid);
    XA_SET_TYPE(&msg.attr, XFS_FILE_REG);
    msg.mode = 0;		/* XXX */
    msg.cred.uid = current->uid;
    msg.cred.pag = xfs_get_pag();
    error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (error == 0)
	error = ((struct xfs_message_wakeup *) & msg)->error;

    if (error == 0)
	error = xfs_lookup(dir, name, len, result);

    return error;
    
}

int
xfs_unlink (struct inode * dir, const char * name, int len)
{
    struct xfs_message_remove msg;
    struct xfs *xfsp  = XFS_FROM_VNODE(dir);
    struct xfs_node *xn = VNODE_TO_XNODE(dir);
    int error;
    
    XFSDEB(XDEBVNOPS, ("xfs_remove: (%s, %d)\n", name, len));
    
    msg.header.opcode = XFS_MSG_REMOVE;
    msg.parent_handle = xn->handle;
    strncpy(msg.name, name, 256);
    msg.cred.uid = current->uid;
    msg.cred.pag = xfs_get_pag();
    error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (error == 0)
        error = ((struct xfs_message_wakeup *) &msg)->error;
    
    iput(dir);
    return error;
}

int
xfs_rename (struct inode * old_dir, const char * old_name, int old_len,
	     struct inode * new_dir, const char * new_name, int new_len,
	     int must_be_dir)
{
    struct xfs *xfsp = XFS_FROM_VNODE(old_dir);
    struct xfs_message_rename msg;
    int error;

    if (old_dir->i_sb != new_dir->i_sb)
	return -EXDEV;
    
    
    msg.header.opcode = XFS_MSG_RENAME;
    msg.old_parent_handle = VNODE_TO_XNODE(old_dir)->handle;
    strncpy(msg.old_name, old_name, 256);
    msg.new_parent_handle = VNODE_TO_XNODE(new_dir)->handle;
    strncpy(msg.new_name, new_name, 256);
    msg.cred.uid = current->uid;
    msg.cred.pag = xfs_get_pag();
    error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
    if (error == 0)
	error = ((struct xfs_message_wakeup *) &msg)->error;

    iput(old_dir);
    iput(new_dir);
    
    return error;
}

static int xfs_readdir(struct inode * inode, struct file * file,
		       void * dirent, filldir_t filldir)
{
    int error = 0;
    int filldir_error;
    off_t offset, begin_offset;
    char *buf;
    long oldfs;

    XFSDEB(XDEBREADDIR, ("xfs_readdir\n"));
    
    error = xfs_data_valid(inode, XFS_DATA_R);
    if (error == 0) {
	struct inode *t = DATA_FROM_VNODE(inode);
	t->i_count++;
	
	while (file->f_pos < t->i_size && error >= 0) {
	  XFSDEB(XDEBREADDIR, ("xfs_readdir file->f_pos: %d t->i_size: %d\n",(int) file->f_pos,(int) t->i_size));
	    if (t->i_op &&
		t->i_op->default_file_ops &&
		t->i_op->default_file_ops->read) {
		buf = xfs_alloc(XFS_DIRENT_BLOCKSIZE);
		begin_offset = file->f_pos &~ (XFS_DIRENT_BLOCKSIZE - 1);
		offset = file->f_pos & (XFS_DIRENT_BLOCKSIZE - 1);
		
		XFSDEB(XDEBREADDIR, ("xfs_readdir begin_offset: %d offset: %d\n",(int)begin_offset,(int)offset));
		oldfs = get_fs();
		set_fs(KERNEL_DS);
		file->f_pos = begin_offset;
		file->f_reada = 0;
		error = t->i_op->default_file_ops->read(t, file, buf, XFS_DIRENT_BLOCKSIZE);
		set_fs(oldfs);
		if (error == XFS_DIRENT_BLOCKSIZE) {
		  XFSDEB(XDEBREADDIR, ("xfs_readdir error: %d\n",error));
		    while (offset < XFS_DIRENT_BLOCKSIZE) {
			XFSDEB(XDEBREADDIR,
			       ("xfs_readdir offset: %d namlen: %d offset2: %d\n",
				(int) offset,
				(int) ((struct xfs_dirent *) (buf + offset))->d_namlen,
				(int) (offset+begin_offset)));
			if ((filldir_error = filldir (dirent,
						      ((struct xfs_dirent *) (buf + offset))->d_name,
						      ((struct xfs_dirent *) (buf + offset))->d_namlen,
						      offset+begin_offset,
						      ((struct xfs_dirent *) (buf + offset))->d_fileno)) < 0) {
			    XFSDEB(XDEBREADDIR, ("xfs_readdir filldir: %d\n", filldir_error));
			    file->f_pos = offset + begin_offset;
			    file->f_reada = 0;
			    iput(t);
			    xfs_free(buf);
			    return 0;
			}
			offset += ((struct xfs_dirent *) (buf + offset))->d_reclen;
		    }
		    xfs_free(buf);
		} else if (error > 0) {
		    printk("XFS Panic! expected %d bytes, got %d in xfs_readdir\n", XFS_DIRENT_BLOCKSIZE, error);
		    error = -EINVAL;
		}
	    } else {
		error = -EINVAL;
		printk("XFS Panic! read not defined in cache file system\n");
	    }
	}
	iput(t);
    }
    
    return error;
}

static int
xfs_readlink(struct inode *inode, char *buffer, int buflen)
{
  int error = 0;
  int fd;
  struct file *file;

  XFSDEB(XDEBVNOPS, ("xfs_readlink\n"));

  error = xfs_data_valid(inode, XFS_DATA_R);
  if (error == 0)
  {
      off_t size;
      long oldfs;

      struct inode *t = DATA_FROM_VNODE(inode);
      if (t == NULL)
	  return -EINVAL;
      fd = open_inode (t, 0);
      if (fd < 0) {
	  printk("XFS Panic! cannot open file\n");
	  return fd;
      }
      file = current->files->fd[fd];
      if (!file) {
	  printk("XFS Panic! cannot fget open file\n");
	  iput(t);
	  return -EINVAL;
      }
      size = (buflen > inode->i_size) ?
	(inode->i_size) :
	(buflen);
      XFSDEB(XDEBVNOPS, ("xfs_readlink buflen: %d inode->i_size: %ud size: %ud\n",buflen, (unsigned int) inode->i_size, (unsigned int) size));
      if (t->i_op &&
	  t->i_op->default_file_ops &&
	  t->i_op->default_file_ops->read) {
	  oldfs = get_fs();
	  set_fs(KERNEL_DS);
	  error = t->i_op->default_file_ops->read(t, file, buffer, size);
	  set_fs(oldfs);
      } else {	
	  printk("XFS Panic! read not defined in cache file system\n");
	  error = -EINVAL;
      }
      sys_close(fd);
  }

  return error;
}

static int xfs_follow_link(struct inode * dir, struct inode * inode,
			   int flag, int mode, struct inode ** res_inode)
{
    int error;
    char *buffer;

    *res_inode = NULL;
    if (dir == NULL) {
	dir = current->fs->root;
	dir->i_count++;
    }
    if (!inode) {
	iput(dir);
	return -ENOENT;
    }
    if (!S_ISLNK(inode->i_mode)) {
	iput(dir);
	*res_inode = inode;
	return 0;
    }
    if (current->link_count > 5) {
	iput(dir);
	iput(inode);
	return -ELOOP;
    }
    buffer = xfs_alloc(MAXPATHLEN + 1);
    if (buffer == NULL) {
	iput(dir);
	iput(inode);
	return -ENOMEM;
    }
    error = xfs_readlink(inode, buffer, MAXPATHLEN);
    if (error < 0) {
	iput(dir);
	iput(inode);
	xfs_free(buffer);
	return error;
    }
    if (error <= MAXPATHLEN)
	buffer[error]='\0';
    current->link_count++;
    error = open_namei (buffer, 0, 0, res_inode, dir);
    current->link_count--;
    iput(inode);
    return error;
}

static void
xfs_release_file (struct inode *inode, struct file *file)
{
    struct xfs *xfsp = XFS_FROM_VNODE(inode);
    struct xfs_node *xn = VNODE_TO_XNODE(inode);
    int error = 0;
    
    XFSDEB(XDEBVNOPS, ("xfs_close\n"));
    
    if (file->f_mode & FMODE_WRITE && xn->flags & XFS_DATA_DIRTY)
	error = do_fsync(xfsp, xn, XFS_WRITE);
    
    if (error)
	printk("xfs_release_file error: %d\n",error);
    
    return;

}
